/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: /usr/src/mash/repository/mash/mash-1/codec/decoder-ldct.cc,v 1.4 1997/11/07 03:51:09 elan Exp $";
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include "config.h"
#include "decoder.h"
#include "dct.h"
#include "ldct-ht.h"
#include "pktbuf.h"

#define LDCT_NLAYER 3

/*
 * True iff there are bits left in the bit-buffer
 * in the indicated level data structure.
 */
#define BB_EOB(lp) ((((lp)->es - (lp)->bs) << 3) + (lp)->nbb <= (lp)->ebit)
#define BB_NBIT(lp) ((((lp)->es - (lp)->bs) << 3) + (lp)->nbb - (lp)->ebit)

#define NRB 64
struct rbent {
	/*XXX packet buffers should be loaned out*/
	u_int16_t seqno;
	u_int16_t flags;
	int next_layer;
	int blkno;
	int qd;
	int S;
	int E;
	int ebit;
	int cc;
	u_char data[RTP_MTU];
	rbent* next;
};

struct pkt_layer {
	/* bit decoding state */
	int nbb;
	u_int bb;
	const u_char* bs;
	const u_char* es;
	int ebit;
	int off;		/* current packet index */

	/* packet reassembly state */
	int next_layer;
	int first;
	int last;
	struct rbent* pkts[NRB]; /* indexed by modulo seqno */
	pkt_layer* next;
};

struct pkt_group {
	u_int32_t ts;/*XXX use this*/
	int group;		/* group ID */
	int base;		/* level of spatial base layer */
	pkt_layer* layers[NLAYER];
};

#define MASK(s) ((1 << (s)) - 1)

/*
 * Read the next 16 bits off the bit string into the bit buffer.
 */
#define READBITS(bb, inb) \
 { \
	register int v__; \
	register const u_char *cp__ = inb; \
 \
	bb <<= 16; \
	v__ = *cp__++; \
	bb |= v__ << 8; \
	v__ = *cp__++; \
	bb |= v__; \
	inb = cp__; \
 \
}

#define GET_BITS(n, nbb, bb, inb, result) \
{ \
	nbb -= n; \
	if (nbb < 0)  { \
		READBITS(bb, inb); \
		nbb += 16; \
	} \
	(result) = ((bb >> nbb) & MASK(n)); \
}

/* Optimize commmon case */
#define GET_BIT(nbb, bb, inb, result) \
{ \
	nbb--; \
	if (nbb < 0)  { \
		READBITS(bb, inb); \
		nbb += 16; \
	} \
	(result) = ((bb >> nbb) & 1); \
}


#define HUFF_DECODE(ht, maxlen, nbb, bb, inb, result) { \
	register int s__, v__; \
 \
	 /*XXX*/ \
	while (nbb < maxlen) { \
		bb <<= 8; \
		bb |= *inb++; \
		nbb += 8; \
	} \
	s__ = maxlen; \
	v__ = (bb >> (nbb - s__)) & MASK(s__); \
	s__ = (ht)[v__]; \
	nbb -= (s__ & 0xffff); \
	result = s__ >> 16; \
}

#define LDCT_GET_BITS(n, lp, v) \
	GET_BITS(n, (lp)->nbb, (lp)->bb, (lp)->bs, v)

#define LDCT_GET_BIT(lp, v) \
	GET_BIT((lp)->nbb, (lp)->bb, (lp)->bs, v)

class LDCTDecoder : public Decoder {
 public:
	LDCTDecoder();
	~LDCTDecoder();
	virtual void recv(pktbuf* pb);
	virtual int command(int argc, const char*const* argv);
	void resize(int width, int height);
	int colorhist(u_int* hist) const;
 protected:
	void decode_group(pkt_group* p);
	void decode_blk(int x, int y);
	void decode_color(u_char* out, int stride);
	virtual void redraw();
	void sync();
	
	int lasteob_;

	pkt_layer* group_;     /* linked list of all layers to make up group */

	u_int8_t* frm_;

	/*XXX*/
	int maxlayer_;

#define NGROUP 4
	pkt_group groups_[NGROUP];

	pkt_layer* alloc_layer();
	void free_layer(pkt_layer*);
	void free_rb(rbent* rb);
	rbent* alloc_rb();
	pkt_layer* free_layer_;
	rbent* free_rb_;

	int add_packet(pkt_group* grp, rbent* rb, int layer);
	void reset_group(pkt_group* p);
	int layer_complete(pkt_layer* p);
	int group_complete(pkt_group* p);

	void init_bbs();
	void check_bbs();
};

static class LDCTDecoderClass : public TclClass {
public:
	LDCTDecoderClass() : TclClass("Module/VideoDecoder/LDCT") {}
	TclObject* create(int, const char*const*) {
		return (new LDCTDecoder());
	}
} swd_ldct_;

struct ldct_layer_spec {
	const int* ht;	/* Huffman table address */
	int cw;		/* Max huff code width */
	int lsb;	/* Least significant bit of layer */
	int len;	/* Length of layer (number of bits) */
};

/* XXX */
const ldct_layer_spec lspec[LDCT_NLAYER] = {
	{ htd_ldct_layer0, htd_ldct_layer0_width, LDCT_L0, LDCT_L0_LEN },
	{ htd_ldct_layer1, htd_ldct_layer1_width, LDCT_L1, LDCT_L1_LEN },
	{ htd_ldct_layer2, htd_ldct_layer2_width, LDCT_L2, LDCT_L2_LEN },
};

#define STAT_BAD_SYNTAX	0
#define STAT_BAD_GEOM	1

LDCTDecoder::LDCTDecoder() : Decoder(sizeof(ldcthdr)),
        frm_(0), maxlayer_(4),
	free_layer_(0), free_rb_(0)
{
	stat_[STAT_BAD_SYNTAX].name = "LDCT-Bad-Syntax";
	stat_[STAT_BAD_GEOM].name = "LDCT-Bad-Geom";
	nstat_ = 2;
	decimation_ = 411;

	/*XXX*/
	inw_ = 320;
	inh_ = 240;

	resize(inw_, inh_);

	memset(groups_, 0, sizeof(groups_));
	for (int i = 0; i < NGROUP; ++i) {
		groups_[i].group = -1;
		groups_[i].base = -1;
	}
}

LDCTDecoder::~LDCTDecoder()
{
	delete frm_;
}

int LDCTDecoder::command(int argc, const char*const* argv)
{
	if (argc == 3) {
		if (strcmp(argv[1], "maxlayer") == 0) {
			maxlayer_ = atoi(argv[2]);
			return (TCL_OK);
		}
	}
	return (Decoder::command(argc, argv));
}

void LDCTDecoder::resize(int w, int h)
{
	delete frm_;
	int size = w*h;

	/* XXX 4:1:1 */
	frm_ = new u_char[2 * size];
	
	/*
	 * Initialize to gray.
	 */
	memset(frm_, 0x80, 2 * size);

	inw_ = w;
	inh_ = h;
	
	Decoder::resize(w, h);
}

extern u_char MSB[];

#define ONECOEFF(c, nbb, bb, inb, rlen, ht, codewidth, len, shift) \
{ \
	if (rlen < len) { \
		u_int ref; \
		int post = len - rlen - 1; \
		int nb = post; \
		int sign; \
		GET_BIT(nbb, bb, inb, sign); /* XXX */ \
		short t__ = 1 << post + shift;	/* MSB */ \
		t__ |= sign << 14;	/* Sign bit */ \
		t__ += 2;		/* amortize low bit quantization */ \
 \
		if (nb > 0) { \
			GET_BITS(nb, nbb, bb, inb, ref); \
			t__ |= ref << shift; /* Ref bits in this layer */ \
		} \
		(c) = t__; \
		/* Read next run length */ \
		HUFF_DECODE(ht, codewidth, nbb, bb, inb, rlen); \
		if (rlen == LDCT_ESC) { \
			GET_BITS(LDCT_ACMAXCODE, nbb, bb, inb, rlen); \
		} else if (rlen == LDCT_EOB) { \
			break; \
		} \
	} else \
		rlen -= len; \
}

#define REFINE(c, nbb, bb, inb, len, shift) \
{ \
	if (c != 0 && MSB[c & ~(1<<14)] >= shift + len) { \
		u_int ref__; \
		GET_BITS(len, nbb, bb, inb, ref__); \
		c |= ref__ << shift; \
	} \
}

/* 
 * Convert magnitudes to 2's-complement in the last layer refinement and 
 * set mask for rdct.
 */
#define LASTREFINE(c, nbb, bb, inb, len, shift, mask, m) \
{\
	if (c != 0) { \
		short c__ = c; \
		if (MSB[c__ & ~(1<<14)] >= shift + len) { \
			u_int ref__; \
			GET_BITS(len, nbb, bb, inb, ref__); \
			c__ |= ref__ << shift; \
		} \
		if ((c__ & (1 << 14)) != 0) \
			c = -(c__ & ~(1 << 14)); \
		else \
			c = c__; \
		(m) |= mask; \
	} \
}

#define L_ONELAYER(b0, b1, b2, b3, nbb0, bb0, inb0, shift, len, ht, \
		   codewidth) \
{ \
	register u_int bb__ = bb0; \
	register int nbb__ = nbb0; \
	register const u_char* inb__ = inb0; \
 \
	int eob__ = -1; \
	int coeff__; \
	int rlen__; \
	HUFF_DECODE(ht, codewidth, nbb__, bb__, inb__, rlen__); \
	if (rlen__ == LDCT_ESC) { \
		GET_BITS(LDCT_ACMAXCODE, nbb__, bb__, inb__, rlen__); \
	} \
	if (rlen__ != LDCT_EOB) { \
		for (coeff__ = 1; coeff__ < 64; coeff__++) { \
			eob__ = coeff__; \
			int z__ = COLZAG[coeff__]; \
			if (b0[z__] == 0) { \
				ONECOEFF(b0[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b1[z__] == 0) { \
				ONECOEFF(b1[z__], nbb__, bb__, inb__,  \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b2[z__] == 0) { \
				ONECOEFF(b2[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b3[z__] == 0) { \
				ONECOEFF(b3[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
		} \
	} \
 \
	for (coeff__ = 1; coeff__ <= lasteob_; coeff__++) { \
		int z__ = COLZAG[coeff__]; \
		REFINE(b0[z__], nbb__, bb__, inb__, len, shift); \
		REFINE(b1[z__], nbb__, bb__, inb__, len, shift); \
		REFINE(b2[z__], nbb__, bb__, inb__, len, shift); \
		REFINE(b3[z__], nbb__, bb__, inb__, len, shift); \
	} \
 \
        if (eob__ > lasteob_) \
		lasteob_ = eob__; \
 \
	bb0 = bb__; \
	nbb0 = nbb__; \
	inb0 = inb__; \
}

#define L_LASTLAYER(b0, b1, b2, b3, nbb0, bb0, inb0, shift, len, ht, \
		    codewidth, bv0, bv1, bv2, bv3) \
{ \
	register u_int bb__ = bb0; \
	register int nbb__ = nbb0; \
	register const u_char* inb__ = inb0; \
 \
 	int eob__ = -1; \
	int coeff__; \
	int rlen__ ; \
	HUFF_DECODE(ht, codewidth, nbb__, bb__, inb__, rlen__); \
	if (rlen__ == LDCT_ESC) { \
		GET_BITS(LDCT_ACMAXCODE, nbb__, bb__, inb__, rlen__); \
	} \
	if (rlen__ != LDCT_EOB) { \
		for (coeff__ = 1; coeff__ < 64; coeff__++) { \
			eob__ = coeff__; \
			int z__ = COLZAG[coeff__]; \
			if (b0[z__] == 0) { \
				ONECOEFF(b0[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b1[z__] == 0) { \
				ONECOEFF(b1[z__], nbb__, bb__, inb__,  \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b2[z__] == 0) { \
				ONECOEFF(b2[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
			if (b3[z__] == 0) { \
				ONECOEFF(b3[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
		} \
	} \
 \
	 /* Need to set lasteob_ to last non-zero coefficient */ \
        if (eob__ > lasteob_) \
		lasteob_ = eob__; \
 \
	(bv0)[0] = (bv1)[0] = (bv2)[0] = (bv3)[0] = 1; \
	(bv0)[1] = (bv1)[1] = (bv2)[1] = (bv3)[1] = 0; \
	for (coeff__ = 1; coeff__ <= lasteob_; coeff__++) { \
		int z__ = COLZAG[coeff__]; \
		if (z__ >= 32) { \
			u_int32_t mask__ = 1 << z__ - 32; \
			LASTREFINE(b0[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv0)[1]); \
			LASTREFINE(b1[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv1)[1]); \
			LASTREFINE(b2[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv2)[1]); \
			LASTREFINE(b3[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv3)[1]); \
		} else { \
			u_int32_t mask__ = 1 << z__; \
			LASTREFINE(b0[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv0)[0]); \
			LASTREFINE(b1[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv1)[0]); \
			LASTREFINE(b2[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv2)[0]); \
			LASTREFINE(b3[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv3)[0]); \
		}\
	} \
 \
	bb0 = bb__; \
	nbb0 = nbb__; \
	inb0 = inb__; \
}

#define C_ONELAYER(b0, nbb0, bb0, inb0, shift, len, ht, codewidth) \
{ \
	register u_int bb__ = bb0; \
	register int nbb__ = nbb0; \
	register const u_char* inb__ = inb0; \
 \
	int eob__ = -1; \
	int coeff__; \
	int rlen__; \
	HUFF_DECODE(ht, codewidth, nbb__, bb__, inb__, rlen__); \
	if (rlen__ == LDCT_ESC) { \
		GET_BITS(LDCT_ACMAXCODE, nbb__, bb__, inb__, rlen__); \
	} \
	if (rlen__ != LDCT_EOB) { \
		for (coeff__ = 1; coeff__ < 64; coeff__++) { \
			eob__ = coeff__; \
			int z__ = COLZAG[coeff__]; \
			if (b0[z__] == 0) { \
				ONECOEFF(b0[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
		} \
	} \
 \
	for (coeff__ = 1; coeff__ <= lasteob_; coeff__++) { \
		int z__ = COLZAG[coeff__]; \
		REFINE(b0[z__], nbb__, bb__, inb__, len, shift); \
	} \
 \
        if (eob__ > lasteob_) \
		lasteob_ = eob__; \
 \
	bb0 = bb__; \
	nbb0 = nbb__; \
	inb0 = inb__; \
}


#define C_LASTLAYER(b0, nbb0, bb0, inb0, shift, len, ht, codewidth, bv0) \
{ \
	register u_int bb__ = bb0; \
	register int nbb__ = nbb0; \
	register const u_char* inb__ = inb0; \
 \
	int eob__ = -1; \
	int coeff__; \
	int rlen__; \
	HUFF_DECODE(ht, codewidth, nbb__, bb__, inb__, rlen__); \
	if (rlen__ == LDCT_ESC) { \
		GET_BITS(LDCT_ACMAXCODE, nbb__, bb__, inb__, rlen__); \
	} \
	if (rlen__ != LDCT_EOB) { \
		for (coeff__ = 1; coeff__ < 64; coeff__++) { \
			/* \
			 * In last pass need to set lasteob to last non zero \
			 * coefficient for last refinement below.  \
			 */ \
			eob__ = coeff__; \
			int z__ = COLZAG[coeff__]; \
			if (b0[z__] == 0) { \
				ONECOEFF(b0[z__], nbb__, bb__, inb__, \
					 rlen__, ht, codewidth, len, shift); \
			} \
		} \
	} \
 \
 	if (eob__ > lasteob_) \
		lasteob_ = eob__; \
 \
	(bv0)[0] = 1; \
	(bv0)[1] = 0; \
	for (coeff__ = 1; coeff__ <= lasteob_; coeff__++) { \
		int z__ = COLZAG[coeff__]; \
		if (z__ >= 32) { \
			u_int32_t mask__ = 1 << z__ - 32; \
			LASTREFINE(b0[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv0)[1]); \
		} else { \
			u_int32_t mask__ = 1 << z__; \
			LASTREFINE(b0[z__], nbb__, bb__, inb__, len, \
				   shift, mask__, (bv0)[0]); \
		}\
	} \
 \
	bb0 = bb__; \
	nbb0 = nbb__; \
	inb0 = inb__; \
}

#define DCDECODE(out, nbb, bb, inb, pred) \
{ \
	int c__; \
	HUFF_DECODE(htd_ldct_dchuff, htd_ldct_dchuff_width, nbb, bb, inb, \
		    c__); \
	if (c__ == LDCT_ESC) { \
		GET_BITS(LDCT_DCMAXCODE, nbb, bb, inb, c__); \
	} \
  	c__ -= 256; \
	(out) = (((pred) + c__) << 3) + 4; \
	(pred) += c__; \
}


/*
 * Luminance and chrominance layer macro hooks 
 */
#define LDCT_L_ONELAYER(lp, ls, b0, b1, b2, b3) \
	  L_ONELAYER((b0), (b1), (b2), (b3), (lp)->nbb, (lp)->bb, (lp)->bs, \
		   (ls)->lsb, (ls)->len, (ls)->ht, (ls)->cw)

#define LDCT_L_LASTLAYER(lp, ls, b0, b1, b2, b3, bv0, bv1, bv2, bv3) \
	  L_LASTLAYER((b0), (b1), (b2), (b3), (lp)->nbb, (lp)->bb, (lp)->bs, \
		      (ls)->lsb, (ls)->len, (ls)->ht, (ls)->cw, \
		      (bv0), (bv1), (bv2), (bv3))

#define LDCT_C_ONELAYER(lp, ls, b0) \
	  C_ONELAYER((b0), (lp)->nbb, (lp)->bb, (lp)->bs, \
		   (ls)->lsb, (ls)->len, (ls)->ht, (ls)->cw)

#define LDCT_C_LASTLAYER(lp, ls, b0, bv0) \
	  C_LASTLAYER((b0), (lp)->nbb, (lp)->bb, (lp)->bs, (ls)->lsb, \
		      (ls)->len, (ls)->ht, (ls)->cw, \
		      (bv0))

void LDCTDecoder::decode_blk(int x, int y)
{
	short blk[256];
	bzero(blk, 256 * sizeof(short));

	short* b0 = blk;
	short* b1 = b0 + 64;
	short* b2 = b1 + 64;
	short* b3 = b2 + 64;
	
	/* 
	 * DC Coefficients
	 */
	register pkt_layer* lp = group_;
	int nbb = lp->nbb;
	u_int bb = lp->bb;
	const u_char* inb = lp->bs;
	int pred;
	GET_BITS(8, nbb, bb, inb, pred);
	b0[0] = (pred << 3) + 4;
	DCDECODE(b1[0], nbb, bb, inb, pred);
	DCDECODE(b2[0], nbb, bb, inb, pred);
	DCDECODE(b3[0], nbb, bb, inb, pred);
	lp->nbb = nbb;
	lp->bb = bb;
	lp->bs = inb;
		
	/* 
	 * AC Coefficients 
	 */
	lasteob_ = -1;

	u_int32_t bv[8];
	for (lp = group_; lp != 0; lp = lp->next) {
		u_int qd = lp->pkts[lp->off & (NRB-1)]->qd;
		
		int bl = LDCT_QD_BL(qd);
		int rlnum = LDCT_QD_RLNUM(qd);
		register int top_layer = bl + rlnum;

		int sl;
		const ldct_layer_spec* ls;
		for (sl = bl; sl < top_layer; ++sl) {
			ls = &lspec[sl];
			LDCT_L_ONELAYER(lp, ls, b0, b1, b2, b3);
		}

		ls = &lspec[top_layer];
		if (LDCT_QD_LASTLAYER(qd) || lp->next_layer > maxlayer_) {
			LDCT_L_LASTLAYER(lp, ls, b0, b1, b2, b3, 
					 bv, bv+2, bv+4, bv+6);
		} else {
			LDCT_L_ONELAYER(lp, ls, b0, b1, b2, b3);
		}
	}

	int off = y * inw_ + x;
	u_int8_t* out = frm_ + off;

	rdct(b0, bv[0], bv[1], out, inw_, (const u_char*)0); 
	rdct(b1, bv[2], bv[3], out+8, inw_, (const u_char*)0); 
	out += inw_ << 3;
	rdct(b2, bv[4], bv[5], out, inw_, (const u_char*)0); 
	rdct(b3, bv[6], bv[7], out + 8, inw_, (const u_char*)0); 
	
	int s = inw_ * inh_;
	u_int8_t* chm = frm_ + s;
	chm += (y >> 1) * (inw_ >> 1) + (x >> 1);
	decode_color(chm, inw_ >> 1);
	decode_color(chm + (s >> 2), inw_ >> 1);
}

void LDCTDecoder::decode_color(u_int8_t* out, int stride)
{
	short blk[64];
	bzero(blk, 64 * sizeof(short));

	register short* b0 = blk;
	register pkt_layer* lp = group_;

	/* 
	 * DC Coefficient
	 */
	int pred;
	GET_BITS(8, lp->nbb, lp->bb, lp->bs, pred);
	b0[0] = (pred << 3) + 4;

	/* 
	 * AC Coefficients 
	 */
	lasteob_ = -1;

	u_int32_t bv[2];
	for (lp = group_; lp != 0; lp = lp->next) {
		u_int qd = lp->pkts[lp->off & (NRB-1)]->qd;
		
		int bl = LDCT_QD_BL(qd);
		int rlnum = LDCT_QD_RLNUM(qd);
		register int top_layer = bl + rlnum;

		int sl;
		const ldct_layer_spec* ls;
		for (sl = bl; sl < top_layer; ++sl) {
			ls = &lspec[sl];
			LDCT_C_ONELAYER(lp, ls, b0);
		}
			
		ls = &lspec[top_layer];
		if (LDCT_QD_LASTLAYER(qd) || lp->next_layer > maxlayer_) {
			LDCT_C_LASTLAYER(lp, ls, b0, bv);
		} else {
			LDCT_C_ONELAYER(lp, ls, b0);
		}
	}

	rdct(b0, bv[0], bv[1], out, stride, (const u_char*)0); 
}

void LDCTDecoder::init_bbs()
{
	for (pkt_layer* p = group_; p != 0; p = p->next) {
		p->nbb = 0;
		p->bb = 0;
		p->off = p->first;
		rbent* rb = p->pkts[p->first & (NRB-1)];
		p->bs = &rb->data[0];
		p->es = &rb->data[rb->cc];
		p->ebit = rb->ebit;
	}
}


/*
 * go through each layer and see if bit buffer has ran out
 * of bits.  this can only happen on block boundaries.
 */
void LDCTDecoder::check_bbs()
{
	for (pkt_layer* p = group_; p != 0; p = p->next) {
#ifdef notdef
		const u_char* cp = p->bs - (p->nbb >> 3);
#endif
		if (BB_EOB(p)) {
			p->nbb = 0;
			p->bb = 0;
			int k = p->off;
			if (k != p->last) {
				k = (k + 1) & 0xffff;
				p->off = k;
				rbent* rb = p->pkts[k & (NRB-1)];
				p->bs = &rb->data[0];
				p->es = &rb->data[rb->cc];
				p->ebit = rb->ebit;
			} else {
				/*XXX*/
				p->bs = 0;
			}
		}
	}
}

void LDCTDecoder::decode_group(pkt_group* p)
{
	/*
	 * link up the possibly discontiguous layers into
	 * a single linked list whose head is "group_"
	 */
	pkt_layer** pp = &group_;
	int i = p->base;
	while (i <= maxlayer_) {
		pkt_layer* lp = p->layers[i];
		*pp = lp;
		pp = &lp->next;
		i = lp->next_layer;
		if (i == 0)
			break;
	}
	*pp = 0;

	init_bbs();

	/*XXX*/
	int blkw = inw_ >> 4;

	/*XXX*/
	int blkno = group_->pkts[group_->off & (NRB-1)]->blkno;

	while (!BB_EOB(group_)) {
		check_bbs();/*XXX*/

		int v;
		LDCT_GET_BIT(group_, v);
		if (v == 1)
			blkno += 1;
		else {
			LDCT_GET_BIT(group_, v);
			if (v == 1) {
				LDCT_GET_BITS(4, group_, v);
				v += 2;
			} else {
				LDCT_GET_BITS(11, group_, v);
			}
			blkno += v;
		}
		int y = blkno / blkw;
		int x = blkno - y * blkw;

		x <<= 4;
		y <<= 4;
		if ((unsigned)x > (unsigned)inw_ ||
		    (unsigned)y > (unsigned)inh_) {
			count(STAT_BAD_SYNTAX);
			return;
		}
		decode_blk(x, y);

		/*XXX*/
		int k = (x >> 3) + (inw_ >> 3) * (y >> 3);
		rvts_[k] = now_;
		rvts_[k + 1] = now_;
		k += inw_ >> 3;
		rvts_[k] = now_;
		rvts_[k + 1] = now_;
	}
}

#define SEQ_LEQ(a, b) (((a) - (b)) << 16 >> 16 <= 0)
#define SEQ_GEQ(a, b) ((((a) - (b)) >> 16) == 0)
#define SEQ_LT(a, b)  ((((a) - (b)) >> 16) != 0)

int LDCTDecoder::add_packet(pkt_group* grp, rbent* rb, int layer)
{
	if (LDCT_QD_BL(rb->qd) == 0)
		grp->base = layer;

#ifdef notdef
	if (LDCT_QD_LASTLAYER(rb->qd) != 0)
		grp->lastlayer = layer;
#endif

	if (layer >= NLAYER)
		return (-1);

	/*
	 * make sure next_layer points to a higher layer or is 0.
	 */
	if (rb->next_layer <= layer && rb->next_layer != 0)
		return (-1);

	int seqno = rb->seqno;
	/*XXX range-check layer*/
	pkt_layer* lp = grp->layers[layer];
	if (lp == 0) {
		lp = alloc_layer();
		grp->layers[layer] = lp;
		lp->first = lp->last = seqno;
	} else {
		if (SEQ_GEQ(seqno, lp->last))
			lp->last = seqno;
		else if (SEQ_LT(seqno, lp->first))
			lp->first = seqno;
	}
	/*XXX*/
	lp->next_layer = rb->next_layer;
	int k = seqno & (NRB-1);
	rbent* p = lp->pkts[k];
	if (p != 0)
		/* duplicate? should've been filtered by RTP layer */
		free_rb(p);

	lp->pkts[k] = rb;

	return (0);
}

pkt_layer* LDCTDecoder::alloc_layer()
{
	pkt_layer* p = free_layer_;
	if (p == 0) {
		p = new pkt_layer;
		memset(p, 0, sizeof(*p));
	} else
		/* packet array guaranteed to be null */
		free_layer_ = p->next;
	memset(p, 0, sizeof(*p));
	p->first = p->last = -1;
	return (p);
}

void LDCTDecoder::free_layer(pkt_layer* p)
{
	/* free packets*/
	int s = p->first;
	if (s >= 0) {
		int last = p->last;
		while (SEQ_LEQ(s, last)) {
			int k = s & (NRB-1);
			rbent* rb = p->pkts[k];
			if (rb != 0) {
				free_rb(rb);
				p->pkts[k] = 0;
			}
			s = (s + 1) & 0xffff;
		}
	}
	/* free layer data structure */
	p->next = free_layer_;
	free_layer_ = p;
}

void LDCTDecoder::free_rb(rbent* rb)
{
	rb->next = free_rb_;
	free_rb_ = rb;
}

rbent* LDCTDecoder::alloc_rb()
{
	rbent* rb = free_rb_;
	if (rb != 0)
		free_rb_ = rb->next;
	else
		rb = new rbent;

	return (rb);
}

void LDCTDecoder::reset_group(pkt_group* p)
{
	p->group = -1;
	p->base = -1;
	for (int i = 0; i < NLAYER; ++i) {
		pkt_layer* lp = p->layers[i];
		if (lp != 0) {
			free_layer(lp);
			p->layers[i] = 0;
		}
	}
}

int LDCTDecoder::layer_complete(pkt_layer* p)
{
	int f = p->first;
	int l = p->last;
	if (p->pkts[f & (NRB-1)]->S == 0 || p->pkts[l & (NRB-1)]->E == 0)
		return (0);
	while (f != l) {
		f = (f + 1) & 0xffff;
		if (p->pkts[f & (NRB-1)] == 0)
			return (0);
	}
	return (1);
}

int LDCTDecoder::group_complete(pkt_group* p)
{
	int n = p->base;
	if (n < 0)
		return (0);
	while (n <= maxlayer_) {
		pkt_layer* lp = p->layers[n];
		if (lp == 0)
			return (0);
		if (!layer_complete(lp))
			return (0);
		n = lp->next_layer;
		if (n == 0)
			break;
	}
	return (1);
}

extern int hackLayerNo;

#ifdef LDCT_DEBUG
static void dump(int layer, rtphdr* rh, ldcthdr* sh)
{
	int flags = ntohs(rh->rh_flags);
	printf("[%d] %x %x %c ldct %dx%d S=%d E=%d nl %d grp %d qd %x blk %d\n",
	       layer,
	       ntohl(rh->rh_ts),
	       ntohs(rh->rh_seqno),
	       (flags & RTP_M) ? 'M' : '-',
	       sh->width, sh->height,
	       sh->S, sh->E,
	       sh->nl, sh->group,
	       sh->qd,
	       ntohs(sh->blkno) << 16 >> 16);
}
#endif

void LDCTDecoder::recv(pktbuf *pb)
{
#ifdef notdef
printf("recv layer=%d maxlayer=%d\n", layer, maxlayer_);
#endif

	int layer = pb->layer;

#ifdef LDCT_DEBUG
printf("%d %d\n", layer, maxlayer_);
#endif
	if (layer > maxlayer_)
		return;

	const rtphdr* rh = (rtphdr*)pb->dp;
	const ldcthdr* ph = (ldcthdr*)(rh + 1);
	int cc = pb->len - (sizeof(*rh) + 4);
	
#ifdef LDCT_DEBUG
	dump(layer, rh, ph);
#endif

	int w = ph->width << 3;
	int h = ph->height << 3;
	if (w != inw_ || h != inh_) {
		if ((unsigned)w > 1000 || (unsigned)h > 1000) {
			/*XXX*/
			w = 320;
			h = 240;
			count(STAT_BAD_GEOM);
		}
		resize(w, h);
	}

	/*XXX should loan out buffers from session*/
	rbent* rb = alloc_rb();
	memcpy(rb->data, pb->dp, cc);
	rb->cc = cc;
	rb->seqno = ntohs(rh->rh_seqno);
	rb->qd = ntohs(ph->qd);
	rb->S = ph->S;
	rb->E = ph->E;
	rb->ebit = ph->ebit;
	rb->flags = ntohs(rh->rh_flags);
	rb->next_layer = ph->nl;
	/*XXX*/
	rb->blkno = ntohs(ph->blkno) << 16 >> 16;

	int group = ph->group;
	pkt_group* p = &groups_[group & (NGROUP-1)];
	if (p->group != group) {
		/*
		 * XXX
		 */
#ifdef LDCT_DEBUG
printf("reset group from %d to %d\n", p->group, group);
#endif
		reset_group(p);
		p->group = group;
	}
	add_packet(p, rb, layer);
	if (group_complete(p)) {
#ifdef LDCT_DEBUG
printf("group %d done\n", p->group);
#endif
		/*XXX*/
		pkt_layer* lp = p->layers[p->base];
		int mark = lp->pkts[lp->last & (NRB-1)]->flags & RTP_M;
		decode_group(p);
		reset_group(p);/*XXX*/
		if (mark) {
			render_frame(frm_);
		}
	}
}

/*
 * XXX should use PlaneDecoder class for 411 also
 */

int LDCTDecoder::colorhist(u_int* hist) const
{
	int s = inw_ * inh_;
	colorhist_411_556(hist, frm_, frm_ + s, frm_ + s + (s >> 2),
			  inw_, inh_);
	return (1);
}

void LDCTDecoder::redraw()
{
	Decoder::redraw(frm_);
}
